@@ -16,10 +16,11 @@ from __future__ import annotations
 from datetime import datetime, timedelta
 from dataclasses import dataclass
 import importlib
+import time
 import typing as t
 import warnings
 
-from flask import current_app, g
+from flask import current_app, g, session
 from flask_login import AnonymousUserMixin, LoginManager
 from flask_login import UserMixin as BaseUserMixin
 from flask_login import current_user
@@ -30,6 +31,7 @@ from werkzeug.datastructures import ImmutableList
 from werkzeug.local import LocalProxy
 
 from .babel import FsDomain
+from .change_email import ChangeEmailForm
 from .decorators import (
     default_reauthn_handler,
     default_unauthn_handler,
@@ -112,7 +114,7 @@ if t.TYPE_CHECKING:  # pragma: no cover
     from flask.typing import ResponseValue
     import flask_login.mixins
     from authlib.integrations.flask_client import OAuth
-    from .datastore import Role, User, UserDatastore
+    from .datastore import UserDatastore
 
 
 # List of authentication mechanisms supported.
@@ -134,7 +136,7 @@ _default_config: dict[str, t.Any] = {
     "I18N_DOMAIN": "flask_security",
     "I18N_DIRNAME": "builtin",
     "EMAIL_VALIDATOR_ARGS": None,
-    "PASSWORD_HASH": "bcrypt",
+    "PASSWORD_HASH": "argon2",
     "PASSWORD_SALT": None,
     "PASSWORD_SINGLE_HASH": {
         "django_argon2",
@@ -158,17 +160,18 @@ _default_config: dict[str, t.Any] = {
         # And always last one...
         "plaintext",
     ],
+    "DEPRECATED_PASSWORD_SCHEMES": ["auto"],
     "PASSWORD_HASH_OPTIONS": {},  # Deprecated at passlib 1.7
-    "PASSWORD_HASH_PASSLIB_OPTIONS": {
-        "argon2__rounds": 10  # 1.7.1 default is 2.
-    },  # >= 1.7.1 method to pass options.
+    "PASSWORD_HASH_PASSLIB_OPTIONS": {},  # passlib >= 1.7.1 method to pass options
+    # (as part of CryptoContext.using)
     "PASSWORD_LENGTH_MIN": 8,
     "PASSWORD_COMPLEXITY_CHECKER": None,
     "PASSWORD_CHECK_BREACHED": False,
     "PASSWORD_BREACHED_COUNT": 1,
     "PASSWORD_NORMALIZE_FORM": "NFKD",
     "PASSWORD_REQUIRED": True,
-    "DEPRECATED_PASSWORD_SCHEMES": ["auto"],
+    "HASHING_SCHEMES": ["sha256_crypt", "hex_md5"],
+    "DEPRECATED_HASHING_SCHEMES": ["auto"],
     "LOGIN_URL": "/login",
     "LOGOUT_URL": "/logout",
     "REGISTER_URL": "/register",
@@ -200,6 +203,8 @@ _default_config: dict[str, t.Any] = {
     "REDIRECT_HOST": None,
     "REDIRECT_BEHAVIOR": None,
     "REDIRECT_ALLOW_SUBDOMAINS": False,
+    "REDIRECT_BASE_DOMAIN": None,
+    "REDIRECT_ALLOWED_SUBDOMAINS": [],
     "FORGOT_PASSWORD_TEMPLATE": "security/forgot_password.html",
     "LOGIN_USER_TEMPLATE": "security/login_user.html",
     "REGISTER_USER_TEMPLATE": "security/register_user.html",
@@ -223,6 +228,15 @@ _default_config: dict[str, t.Any] = {
     "SEND_PASSWORD_RESET_EMAIL": True,
     "SEND_PASSWORD_RESET_NOTICE_EMAIL": True,
     "LOGIN_WITHIN": "1 days",
+    "CHANGE_EMAIL": False,
+    "CHANGE_EMAIL_TEMPLATE": "security/change_email.html",
+    "CHANGE_EMAIL_WITHIN": "2 hours",
+    "CHANGE_EMAIL_URL": "/change-email",
+    "CHANGE_EMAIL_CONFIRM_URL": "/change-email-confirm",
+    "CHANGE_EMAIL_ERROR_VIEW": None,  # spa
+    "POST_CHANGE_EMAIL_VIEW": None,  # spa
+    "CHANGE_EMAIL_SALT": "change-email-salt",
+    "CHANGE_EMAIL_SUBJECT": _("Confirm your new email address"),
     "TWO_FACTOR_AUTHENTICATOR_VALIDITY": 120,
     "TWO_FACTOR_MAIL_VALIDITY": 300,
     "TWO_FACTOR_SMS_VALIDITY": 120,
@@ -234,6 +248,8 @@ _default_config: dict[str, t.Any] = {
         "secure": False,
         "samesite": "Strict",
     },
+    "TWO_FACTOR_SETUP_SALT": "tf-setup-salt",
+    "TWO_FACTOR_SETUP_WITHIN": "30 minutes",
     "TWO_FACTOR_RESCUE_EMAIL": True,
     "MULTI_FACTOR_RECOVERY_CODES": False,
     "MULTI_FACTOR_RECOVERY_CODES_N": 5,
@@ -283,9 +299,8 @@ _default_config: dict[str, t.Any] = {
     "PHONE_REGION_DEFAULT": "US",
     "FRESHNESS": timedelta(hours=24),
     "FRESHNESS_GRACE_PERIOD": timedelta(hours=1),
+    "FRESHNESS_ALLOW_AUTH_TOKEN": True,
     "API_ENABLED_METHODS": ["session", "token"],
-    "HASHING_SCHEMES": ["sha256_crypt", "hex_md5"],
-    "DEPRECATED_HASHING_SCHEMES": ["hex_md5"],
     "DATETIME_FACTORY": naive_utcnow,
     "TOTP_SECRETS": None,
     "TOTP_ISSUER": None,
@@ -528,6 +543,10 @@ _default_messages = {
         _("You successfully disabled two factor authorization."),
         "success",
     ),
+    "TWO_FACTOR_SETUP_EXPIRED": (
+        _("Setup must be completed within %(within)s. Please start over."),
+        "error",
+    ),
     "US_CURRENT_METHODS": (
         _("Currently active sign in options: %(method_list)s."),
         "info",
@@ -608,6 +627,21 @@ _default_messages = {
         _("Credential user handle didn't match"),
         "error",
     ),
+    "CHANGE_EMAIL_EXPIRED": (
+        _("Confirmation must be completed within %(within)s. Please start over."),
+        "error",
+    ),
+    "CHANGE_EMAIL_CONFIRMED": (
+        _("Change of email address confirmed"),
+        "success",
+    ),
+    "CHANGE_EMAIL_SENT": (
+        _(
+            "Instructions to confirm your new email address have"
+            " been sent to %(email)s."
+        ),
+        "success",
+    ),
 }
 
 
@@ -650,6 +684,7 @@ def _user_loader(user_id):
     user = _security.datastore.find_user(fs_uniquifier=str(user_id))
     if user and user.active:
         set_request_attr("fs_authn_via", "session")
+        set_request_attr("fs_paa", session.get("fs_paa", 0))
         return user
     return None
 
@@ -683,6 +718,8 @@ def _request_loader(request):
 
     if user and user.active and user.verify_auth_token(tdata):
         set_request_attr("fs_authn_via", "token")
+        if cv("FRESHNESS_ALLOW_AUTH_TOKEN"):
+            set_request_attr("fs_paa", tdata.get("fs_paa", 0))
         return user
 
     return None
@@ -780,9 +817,13 @@ class RoleMixin:
     """Mixin for `Role` model definitions"""
 
     if t.TYPE_CHECKING:  # pragma: no cover
+        id: int
+        name: str
+        description: str | None
+        permissions: list[str] | None
+        update_datetime: datetime
 
-        def __init__(self) -> None:
-            self.permissions: list[str] | None
+        def __init__(self, **kwargs): ...
 
     def __eq__(self, other):
         return self.name == other or self.name == getattr(other, "name", None)
@@ -807,6 +848,35 @@ class RoleMixin:
 class UserMixin(BaseUserMixin):
     """Mixin for `User` model definitions"""
 
+    if t.TYPE_CHECKING:  # pragma: no cover
+        # These are defined in the applications Model files.
+        id: int
+        email: str
+        username: str | None
+        password: str | None
+        active: bool
+        fs_uniquifier: str
+        fs_token_uniquifier: str
+        fs_webauthn_user_handle: str
+        confirmed_at: datetime | None
+        last_login_at: datetime
+        current_login_at: datetime
+        last_login_ip: str | None
+        current_login_ip: str | None
+        login_count: int
+        tf_primary_method: str | None
+        tf_totp_secret: str | None
+        tf_phone_number: str | None
+        mf_recovery_codes: list[str] | None
+        us_phone_number: str | None
+        us_totp_secrets: str | bytes | None
+        create_datetime: datetime
+        update_datetime: datetime
+        roles: list[RoleMixin]
+        webauthn: list[WebAuthnMixin]
+
+        def __init__(self, **kwargs): ...
+
     def get_id(self) -> str:
         """Returns the user identification attribute. 'Alternative-token' for
         Flask-Login. This is always ``fs_uniquifier``.
@@ -825,8 +895,12 @@ class UserMixin(BaseUserMixin):
 
         :raises ValueError: If ``fs_token_uniquifier`` is part of model but not set.
 
-        Optionally use a separate uniquifier so that changing password doesn't
-        invalidate auth tokens.
+        Uses ``fs_uniquifier`` or ``fs_token_uniquifier`` (if in the UserModel)
+        to identify this user. If ``fs_token_uniquifier`` is used then
+        changing password doesn't invalidate auth tokens.
+
+        Calls :meth:`.UserMixin.augment_auth_token` which applications can override
+        to add any additional information.
 
         The returned value is securely signed using the ``remember_token_serializer``
 
@@ -836,6 +910,9 @@ class UserMixin(BaseUserMixin):
         .. versionchanged:: 5.4.0
             New format - a dict with a version string. Add a token-based expiry
             option as well as a session id.
+        .. versionchanged:: 5.5.0
+            Remove session id (never set or used); added fs_paa (last authentication
+            timestamp)
         """
 
         tdata: dict[str, t.Any] = dict(ver=str(5))
@@ -845,7 +922,9 @@ class UserMixin(BaseUserMixin):
             tdata["uid"] = str(self.fs_token_uniquifier)
         else:
             tdata["uid"] = str(self.fs_uniquifier)
-        tdata["sid"] = 0  # session id
+        # Set the primary authenticated at variable. This is equivalent to
+        # what we set in the session.
+        tdata["fs_paa"] = time.time()  # equivalent of session["fs_paa"]
         tdata["exp"] = int(cv("TOKEN_EXPIRE_TIMESTAMP")(self))  # if >0 then shorter of
         # :data:SECURITY_MAX_AGE and this.
 
@@ -857,7 +936,8 @@ class UserMixin(BaseUserMixin):
 
     def augment_auth_token(self, tdata: dict[str, t.Any]) -> None:
         """Override this to add/modify parts of the auth token.
-        Additions to the dict can be made and verified in verify_auth_token()
+        Additions to the dict can be made here and verified in
+        :meth:`.UserMixin.verify_auth_token`
 
         .. versionadded:: 5.4.0
         """
@@ -879,7 +959,7 @@ class UserMixin(BaseUserMixin):
         """
         return True
 
-    def has_role(self, role: str | Role) -> bool:
+    def has_role(self, role: str | RoleMixin) -> bool:
         """Returns `True` if the user identifies with the specified role.
 
         :param role: A role name or `Role` instance"""
@@ -1004,11 +1084,27 @@ class UserMixin(BaseUserMixin):
 
 
 class WebAuthnMixin:
+    if t.TYPE_CHECKING:  # pragma: no cover
+        # These are defined in the applications Model files.
+        id: int
+        name: str
+        credential_id: bytes
+        public_key: bytes
+        sign_count: int
+        transports: list[str] | None
+        backup_state: bool
+        device_type: str
+        extensions: str | None
+        lastuse_datetime: datetime
+        usage: str
+
+        def __init__(self, **kwargs): ...
+
     def get_user_mapping(self) -> dict[str, t.Any]:
         """
         Return the filter needed by find_user() to get the user
         associated with this webauthn credential.
-        Note that this probably has to be overridden using mongoengine.
+        Note that this probably has to be overridden when using mongoengine.
 
         .. versionadded:: 5.0.0
         """
@@ -1034,10 +1130,11 @@ class Security:
     :param register_blueprint: to register the Security blueprint or not.
     :param login_form: set form for the login view
     :param verify_form: set form for re-authentication due to freshness check
+    :param change_email_form: set form for changing email address
     :param register_form: set form for the register view when
-            *SECURITY_CONFIRMABLE* is false
+            :data:`SECURITY_CONFIRMABLE` is false
     :param confirm_register_form: set form for the register view when
-            *SECURITY_CONFIRMABLE* is true
+            :data:`SECURITY_CONFIRMABLE` is true
     :param forgot_password_form: set form for the forgot password view
     :param reset_password_form: set form for the reset password view
     :param change_password_form: set form for the change password view
@@ -1086,7 +1183,8 @@ class Security:
 
     .. versionadded:: 3.4.0
         ``us_signin_form``, ``us_setup_form``, ``us_setup_validate_form``, and
-        ``us_verify_form`` added as part of the :ref:`unified-sign-in` feature.
+        ``us_verify_form`` added as part of the :ref:`configuration:unified signin`
+         feature.
 
     .. versionadded:: 3.4.0
         ``totp_cls`` added to enable applications to implement replay protection - see
@@ -1114,6 +1212,9 @@ class Security:
         ``mf_recovery_form``.
     .. versionadded:: 5.1.0
         ``mf_recovery_codes_util_cls``, ``oauth``
+    .. versionadded:: 5.5.0
+        ``change_email_form`` in support of the
+         :ref:`Change-Email<configuration:change-email>` feature.
 
     .. deprecated:: 4.0.0
         ``send_mail`` and ``send_mail_task``. Replaced with ``mail_util_cls``.
@@ -1132,8 +1233,10 @@ class Security:
         app: flask.Flask | None = None,
         datastore: UserDatastore | None = None,
         register_blueprint: bool = True,
+        *,
         login_form: t.Type[LoginForm] = LoginForm,
         verify_form: t.Type[VerifyForm] = VerifyForm,
+        change_email_form: t.Type[ChangeEmailForm] = ChangeEmailForm,
         confirm_register_form: t.Type[ConfirmRegisterForm] = ConfirmRegisterForm,
         register_form: t.Type[RegisterForm] = RegisterForm,
         forgot_password_form: t.Type[ForgotPasswordForm] = ForgotPasswordForm,
@@ -1207,6 +1310,7 @@ class Security:
             "register_form": FormInfo(cls=register_form),
             "forgot_password_form": FormInfo(cls=forgot_password_form),
             "reset_password_form": FormInfo(cls=reset_password_form),
+            "change_email_form": FormInfo(cls=change_email_form),
             "change_password_form": FormInfo(cls=change_password_form),
             "send_confirmation_form": FormInfo(cls=send_confirmation_form),
             "passwordless_login_form": FormInfo(cls=passwordless_login_form),
@@ -1237,7 +1341,7 @@ class Security:
             default_unauthz_handler
         )
         self._render_json: t.Callable[
-            [dict[str, t.Any], int, dict[str, str] | None, User | None],
+            [dict[str, t.Any], int, dict[str, str] | None, UserMixin | None],
             ResponseValue,
         ] = default_render_json
         self._want_json: t.Callable[[Request], bool] = default_want_json
@@ -1246,8 +1350,10 @@ class Security:
         self.remember_token_serializer: URLSafeTimedSerializer
         self.login_serializer: URLSafeTimedSerializer
         self.reset_serializer: URLSafeTimedSerializer
+        self.change_email_serializer: URLSafeTimedSerializer
         self.confirm_serializer: URLSafeTimedSerializer
         self.us_setup_serializer: URLSafeTimedSerializer
+        self.tf_setup_serializer: URLSafeTimedSerializer
         self.tf_validity_serializer: URLSafeTimedSerializer
         self.wan_serializer: URLSafeTimedSerializer
         self.principal: Principal
@@ -1269,10 +1375,12 @@ class Security:
         self._password_util: PasswordUtil
         self._totp_factory: Totp
         self._username_util: UsernameUtil
+        self._webauthn_util: WebauthnUtil
         self._mf_recovery_codes_util: MfRecoveryCodesUtil
 
         # Add necessary attributes here to keep mypy happy
         self.trackable: bool = False
+        self.change_email: bool = False
         self.confirmable: bool = False
         self.registerable: bool = False
         self.changeable: bool = False
@@ -1339,6 +1447,7 @@ class Security:
         form_names = [
             "login_form",
             "verify_form",
+            "change_email_form",
             "confirm_register_form",
             "register_form",
             "forgot_password_form",
@@ -1374,6 +1483,7 @@ class Security:
         attr_names = [
             "trackable",
             "registerable",
+            "change_email",
             "confirmable",
             "changeable",
             "recoverable",
@@ -1387,6 +1497,7 @@ class Security:
             "render_template",
             "totp_cls",
             "webauthn_util_cls",
+            "datetime_factory",
         ]
         for attr in attr_names:
             if ov := kwargs.get(attr, cv(attr.upper(), app, strict=False)):
@@ -1422,8 +1533,10 @@ class Security:
         self.remember_token_serializer = _get_serializer(app, "remember")
         self.login_serializer = _get_serializer(app, "login")
         self.reset_serializer = _get_serializer(app, "reset")
+        self.change_email_serializer = _get_serializer(app, "change_email")
         self.confirm_serializer = _get_serializer(app, "confirm")
         self.us_setup_serializer = _get_serializer(app, "us_setup")
+        self.tf_setup_serializer = _get_serializer(app, "two_factor_setup")
         self.tf_validity_serializer = _get_serializer(app, "two_factor_validity")
         self.wan_serializer = _get_serializer(app, "wan")
         self.principal = _get_principal(app)
@@ -1634,6 +1747,7 @@ class Security:
 
         # If they don't want ALL mechanisms protected, then they must
         # set WTF_CSRF_CHECK_DEFAULT=False so that our decorators get control.
+        # And our decorators use csrf.protect()
         if cv("CSRF_PROTECT_MECHANISMS", app=app) != AUTHN_MECHANISMS:
             if not csrf:
                 # This isn't good.
@@ -1726,7 +1840,7 @@ class Security:
     def render_json(
         self,
         cb: t.Callable[
-            [dict[str, t.Any], int, dict[str, str] | None, User | None],
+            [dict[str, t.Any], int, dict[str, str] | None, UserMixin | None],
             ResponseValue,
         ],
     ) -> None:
@@ -1895,6 +2009,11 @@ class Security:
     ) -> None:
         self._add_ctx_processor("reset_password", fn)
 
+    def change_email_context_processor(
+        self, fn: t.Callable[[], dict[str, t.Any]]
+    ) -> None:
+        self._add_ctx_processor("change_email", fn)
+
     def change_password_context_processor(
         self, fn: t.Callable[[], dict[str, t.Any]]
     ) -> None:
